Package: hms
Title: Pretty Time of Day
Date: 2023-03-21
Version: 1.1.3
Authors@R: c(
    person("Kirill", "Müller", role = c("aut", "cre"), email = "kirill@cynkra.com", comment = c(ORCID = "0000-0002-1416-3412")),
    person("R Consortium", role = "fnd"),
    person("RStudio", role = "fnd")
    )
Description: Implements an S3 class for storing and formatting time-of-day
    values, based on the 'difftime' class.
Imports: lifecycle, methods, pkgconfig, rlang (>= 1.0.2), vctrs (>=
        0.3.8)
Suggests: crayon, lubridate, pillar (>= 1.1.0), testthat (>= 3.0.0)
License: MIT + file LICENSE
Encoding: UTF-8
URL: https://hms.tidyverse.org/, https://github.com/tidyverse/hms
BugReports: https://github.com/tidyverse/hms/issues
RoxygenNote: 7.2.3
Config/testthat/edition: 3
Config/autostyle/scope: line_breaks
Config/autostyle/strict: false
Config/Needs/website: tidyverse/tidytemplate
NeedsCompilation: no
Packaged: 2023-03-21 16:52:11 UTC; kirill
Author: Kirill Müller [aut, cre] (<https://orcid.org/0000-0002-1416-3412>),
  R Consortium [fnd],
  RStudio [fnd]
Maintainer: Kirill Müller <kirill@cynkra.com>
Repository: CRAN
Date/Publication: 2023-03-21 18:10:02 UTC
Built: R 4.4.0; ; 2026-02-20 16:54:00 UTC; unix

Package: hms
Title: Pretty Time of Day
Date: 2025-10-11
Version: 1.1.4
Authors@R: c(
    person("Kirill", "Müller", , "kirill@cynkra.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-1416-3412")),
    person("R Consortium", role = "fnd"),
    person("Posit Software, PBC", role = "fnd",
           comment = c(ROR = "03wc8by49"))
  )
Description: Implements an S3 class for storing and formatting time-of-day
    values, based on the 'difftime' class.
License: MIT + file LICENSE
URL: https://hms.tidyverse.org/, https://github.com/tidyverse/hms
BugReports: https://github.com/tidyverse/hms/issues
Imports: cli, lifecycle, methods, pkgconfig, rlang (>= 1.0.2), vctrs
        (>= 0.3.8)
Suggests: crayon, lubridate, pillar (>= 1.1.0), testthat (>= 3.0.0)
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3.9000
NeedsCompilation: no
Packaged: 2025-10-16 19:10:16 UTC; kirill
Author: Kirill Müller [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1416-3412>),
  R Consortium [fnd],
  Posit Software, PBC [fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Kirill Müller <kirill@cynkra.com>
Repository: CRAN
Date/Publication: 2025-10-17 05:10:11 UTC
Built: R 4.4.0; ; 2026-07-17 18:25:09 UTC; unix

Package: hms
Title: Pretty Time of Day
Date: 2025-10-11
Version: 1.1.4
Authors@R: c(
    person("Kirill", "Müller", , "kirill@cynkra.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-1416-3412")),
    person("R Consortium", role = "fnd"),
    person("Posit Software, PBC", role = "fnd",
           comment = c(ROR = "03wc8by49"))
  )
Description: Implements an S3 class for storing and formatting time-of-day
    values, based on the 'difftime' class.
License: MIT + file LICENSE
URL: https://hms.tidyverse.org/, https://github.com/tidyverse/hms
BugReports: https://github.com/tidyverse/hms/issues
Imports: cli, lifecycle, methods, pkgconfig, rlang (>= 1.0.2), vctrs
        (>= 0.3.8)
Suggests: crayon, lubridate, pillar (>= 1.1.0), testthat (>= 3.0.0)
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3.9000
NeedsCompilation: no
Packaged: 2025-10-16 19:10:16 UTC; kirill
Author: Kirill Müller [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1416-3412>),
  R Consortium [fnd],
  Posit Software, PBC [fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Kirill Müller <kirill@cynkra.com>
Repository: CRAN
Date/Publication: 2025-10-17 05:10:11 UTC

# Generated by roxygen2: do not edit by hand

S3method("[<-",hms)
S3method("[[",hms)
S3method("units<-",hms)
S3method(as.POSIXct,hms)
S3method(as.POSIXlt,hms)
S3method(as.character,hms)
S3method(as.hms,POSIXlt)
S3method(as.hms,POSIXt)
S3method(as.hms,default)
S3method(as_hms,default)
S3method(c,hms)
S3method(format,hms)
S3method(print,hms)
S3method(unique,hms)
S3method(vec_cast,hms)
S3method(vec_cast.POSIXct,hms)
S3method(vec_cast.POSIXlt,hms)
S3method(vec_cast.character,hms)
S3method(vec_cast.difftime,hms)
S3method(vec_cast.double,hms)
S3method(vec_cast.hms,POSIXct)
S3method(vec_cast.hms,POSIXlt)
S3method(vec_cast.hms,character)
S3method(vec_cast.hms,default)
S3method(vec_cast.hms,difftime)
S3method(vec_cast.hms,double)
S3method(vec_cast.hms,hms)
S3method(vec_cast.hms,integer)
S3method(vec_cast.integer,hms)
S3method(vec_ptype2,hms)
S3method(vec_ptype2.difftime,hms)
S3method(vec_ptype2.hms,default)
S3method(vec_ptype2.hms,difftime)
S3method(vec_ptype2.hms,hms)
S3method(vec_ptype_abbr,hms)
S3method(vec_ptype_full,hms)
export(as.hms)
export(as_hms)
export(hms)
export(is.hms)
export(is_hms)
export(new_hms)
export(parse_hm)
export(parse_hms)
export(round_hms)
export(trunc_hms)
export(vec_cast.hms)
export(vec_ptype2.hms)
import(rlang)
import(vctrs)
importFrom(lifecycle,deprecate_soft)
importFrom(lifecycle,expect_deprecated)
importFrom(methods,setOldClass)
importFrom(pkgconfig,get_config)

# hms <a href='https:/hms.tidyverse.org'><img src='man/figures/logo.png' align="right" height="139" /></a>

<!-- badges: start -->

[![Lifecycle: stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html) [![rcc](https://github.com/tidyverse/hms/workflows/rcc/badge.svg)](https://github.com/tidyverse/hms/actions) [![Codecov test coverage](https://codecov.io/gh/tidyverse/hms/branch/main/graph/badge.svg)](https://app.codecov.io/gh/tidyverse/hms?branch=main) [![CRAN\_Status\_Badge](https://www.r-pkg.org/badges/version/hms)](https://cran.r-project.org/package=hms)

<!-- badges: end -->

## Overview

The hms package provides a simple class for storing durations or time-of-day values and displaying them in the hh:mm:ss format. This class is intended to simplify data exchange with databases, spreadsheets, and other data sources:

-   Stores values as a numeric vector that contains the number of seconds since midnight
-   Supports construction from explicit hour, minute, or second values
-   Supports coercion to and from various data types, including `POSIXt`
-   Can be used as column in a data frame
-   Based on the `difftime` class
-   Values can exceed the 24-hour boundary or be negative
-   By default, fractional seconds up to a microsecond are displayed, regardless of the value of the `"digits.secs"` option

## Installation

<pre class='chroma'>
<span class='c'># The easiest way to get hms is to install the whole tidyverse:</span>
<span class='nf'><a href='https://rdrr.io/r/utils/install.packages.html'>install.packages</a></span><span class='o'>(</span><span class='s'>"tidyverse"</span><span class='o'>)</span>

<span class='c'># Alternatively, install just hms:</span>
<span class='nf'><a href='https://rdrr.io/r/utils/install.packages.html'>install.packages</a></span><span class='o'>(</span><span class='s'>"hms"</span><span class='o'>)</span>

<span class='c'># Or the the development version from GitHub:</span>
<span class='c'># install.packages("devtools")</span>
<span class='nf'>devtools</span><span class='nf'>::</span><span class='nf'><a href='https://devtools.r-lib.org/reference/remote-reexports.html'>install_github</a></span><span class='o'>(</span><span class='s'>"tidyverse/hms"</span><span class='o'>)</span></pre>

## Usage

The following example showcases ways of using the `hms` class standalone or as a data frame column.

<pre class='chroma'>
<span class='kr'><a href='https://rdrr.io/r/base/library.html'>library</a></span><span class='o'>(</span><span class='nv'><a href='https://hms.tidyverse.org/'>hms</a></span><span class='o'>)</span>

<span class='nf'><a href='https://hms.tidyverse.org/reference/hms.html'>hms</a></span><span class='o'>(</span><span class='m'>56</span>, <span class='m'>34</span>, <span class='m'>12</span><span class='o'>)</span>
<span class='c'>#&gt; 12:34:56</span>
<span class='nf'><a href='https://hms.tidyverse.org/reference/hms.html'>as_hms</a></span><span class='o'>(</span><span class='nf'><a href='https://rdrr.io/r/base/Sys.time.html'>Sys.time</a></span><span class='o'>(</span><span class='o'>)</span><span class='o'>)</span>
<span class='c'>#&gt; 11:55:02.553476</span>
<span class='nf'><a href='https://hms.tidyverse.org/reference/parse_hms.html'>parse_hms</a></span><span class='o'>(</span><span class='s'>"12:34:56"</span><span class='o'>)</span>
<span class='c'>#&gt; 12:34:56</span>
<span class='nf'><a href='https://rdrr.io/r/base/as.POSIXlt.html'>as.POSIXct</a></span><span class='o'>(</span><span class='nf'><a href='https://hms.tidyverse.org/reference/hms.html'>hms</a></span><span class='o'>(</span><span class='m'>1</span><span class='o'>)</span><span class='o'>)</span>
<span class='c'>#&gt; [1] "1970-01-01 00:00:01 UTC"</span>

<span class='nf'><a href='https://rdrr.io/r/base/data.frame.html'>data.frame</a></span><span class='o'>(</span>hours <span class='o'>=</span> <span class='m'>1</span><span class='o'>:</span><span class='m'>3</span>, hms <span class='o'>=</span> <span class='nf'><a href='https://hms.tidyverse.org/reference/hms.html'>hms</a></span><span class='o'>(</span>hours <span class='o'>=</span> <span class='m'>1</span><span class='o'>:</span><span class='m'>3</span><span class='o'>)</span><span class='o'>)</span>
<span class='c'>#&gt;   hours      hms</span>
<span class='c'>#&gt; 1     1 01:00:00</span>
<span class='c'>#&gt; 2     2 02:00:00</span>
<span class='c'>#&gt; 3     3 03:00:00</span></pre>

## Internal representation

Objects of the `hms` and its underlying `difftime` classes are stored as number of seconds since `00:00:00`. Use [`as.numeric()`](https://rdrr.io/r/base/numeric.html) and [`as_hms()`](https://hms.tidyverse.org/reference/hms.html) to convert to and from numbers.

<pre class='chroma'>
<span class='nv'>times</span> <span class='o'>&lt;-</span> <span class='nf'><a href='https://hms.tidyverse.org/reference/parse_hms.html'>parse_hms</a></span><span class='o'>(</span><span class='nf'><a href='https://rdrr.io/r/base/c.html'>c</a></span><span class='o'>(</span><span class='s'>"00:00:00.25"</span>, <span class='s'>"00:00:01"</span>, <span class='s'>"00:01:30"</span>, <span class='s'>"01:00:00"</span><span class='o'>)</span><span class='o'>)</span>
<span class='nv'>times</span>
<span class='c'>#&gt; 00:00:00.25</span>
<span class='c'>#&gt; 00:00:01.00</span>
<span class='c'>#&gt; 00:01:30.00</span>
<span class='c'>#&gt; 01:00:00.00</span>
<span class='nv'>times_num</span> <span class='o'>&lt;-</span> <span class='nf'><a href='https://rdrr.io/r/base/numeric.html'>as.numeric</a></span><span class='o'>(</span><span class='nv'>times</span><span class='o'>)</span>
<span class='nv'>times_num</span>
<span class='c'>#&gt; [1]    0.25    1.00   90.00 3600.00</span>
<span class='nf'><a href='https://hms.tidyverse.org/reference/hms.html'>as_hms</a></span><span class='o'>(</span><span class='nv'>times_num</span><span class='o'>)</span>
<span class='c'>#&gt; 00:00:00.25</span>
<span class='c'>#&gt; 00:00:01.00</span>
<span class='c'>#&gt; 00:01:30.00</span>
<span class='c'>#&gt; 01:00:00.00</span></pre>

------------------------------------------------------------------------

Please note that the ‘hms’ project is released with a [Contributor Code of Conduct](https://github.com/tidyverse/hms/blob/master/CODE_OF_CONDUCT.md). By contributing to this project, you agree to abide by its terms.

<!-- README.md is generated from README.Rmd. Please edit that file -->

# hms <a href='https://hms.tidyverse.org'><img src='man/figures/logo.png' align="right" height="139" alt="Hexagonal logo for the R package ‘hms’, featuring a stylized green clock face showing the time 10:05, with the package name ‘hms’ in white text at the center and the rstudio.com URL at the bottom edge."/></a>

<!-- badges: start -->

[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![rcc](https://github.com/tidyverse/hms/workflows/rcc/badge.svg)](https://github.com/tidyverse/hms/actions)
[![Codecov test
coverage](https://codecov.io/gh/tidyverse/hms/branch/main/graph/badge.svg)](https://app.codecov.io/gh/tidyverse/hms?branch=main)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/hms)](https://cran.r-project.org/package=hms)
<!-- badges: end -->

## Overview

The hms package provides a simple class for storing durations or
time-of-day values and displaying them in the hh:mm:ss format. This
class is intended to simplify data exchange with databases,
spreadsheets, and other data sources:

- Stores values as a numeric vector that contains the number of seconds
  since midnight
- Supports construction from explicit hour, minute, or second values
- Supports coercion to and from various data types, including `POSIXt`
- Can be used as column in a data frame
- Based on the `difftime` class
- Values can exceed the 24-hour boundary or be negative
- By default, fractional seconds up to a microsecond are displayed,
  regardless of the value of the `"digits.secs"` option

## Installation

``` r
# The easiest way to get hms is to install the whole tidyverse:
install.packages("tidyverse")

# Alternatively, install just hms:
install.packages("hms")

# Or the the development version from GitHub:
# install.packages("pak")
pak::pak("tidyverse/hms")
```

## Usage

The following example showcases ways of using the `hms` class standalone
or as a data frame column.

``` r
library(hms)

hms(56, 34, 12)
#> 12:34:56
as_hms(Sys.time())
#> 10:27:38.905625
parse_hms("12:34:56")
#> 12:34:56
as.POSIXct(hms(1))
#> [1] "1970-01-01 00:00:01 UTC"

data.frame(hours = 1:3, hms = hms(hours = 1:3))
#>   hours      hms
#> 1     1 01:00:00
#> 2     2 02:00:00
#> 3     3 03:00:00
```

## Internal representation

Objects of the `hms` and its underlying `difftime` classes are stored as
number of seconds since `00:00:00`. Use `as.numeric()` and `as_hms()` to
convert to and from numbers.

``` r
times <- parse_hms(c("00:00:00.25", "00:00:01", "00:01:30", "01:00:00"))
times
#> 00:00:00.25
#> 00:00:01.00
#> 00:01:30.00
#> 01:00:00.00
times_num <- as.numeric(times)
times_num
#> [1]    0.25    1.00   90.00 3600.00
as_hms(times_num)
#> 00:00:00.25
#> 00:00:01.00
#> 00:01:30.00
#> 01:00:00.00
```

------------------------------------------------------------------------

Please note that the ‘hms’ project is released with a [Contributor Code
of
Conduct](https://github.com/tidyverse/hms/blob/master/CODE_OF_CONDUCT.md).
By contributing to this project, you agree to abide by its terms.

# Dynamically exported, see zzz.R
pillar_shaft.hms <- function(x, ...) {
  data <- rep(NA_character_, length(x))

  xx <- decompose(x)
  has_hours <- xx$hours > 0
  highlight_hours <- has_hours
  highlighted <- highlight_hours

  has_minutes <- xx$minute_of_hour > 0
  highlight_minutes <- !highlighted & has_minutes
  highlighted <- highlighted | highlight_minutes

  has_seconds <- xx$second_of_minute > 0
  highlight_seconds <- !highlighted & has_seconds
  highlighted <- highlighted | highlight_seconds

  has_tics <- xx$tics > 0
  highlight_tics <- !highlighted & has_tics

  need_tics <- any(has_tics, na.rm = TRUE)
  need_seconds <- need_tics || any(has_seconds, na.rm = TRUE)
  need_hours <- any(has_hours, na.rm = TRUE)
  need_sign <- any(xx$sign)

  if (need_hours) {
    data_seconds <- paste0(
      if (need_sign) ifelse(xx$sign, "-", " ") else "",
      pillar::style_num(format_hours(xx$hours), xx$sign, highlight_hours),
      pillar::style_subtle(":"),
      pillar::style_num(format_two_digits(xx$minute_of_hour), xx$sign, highlight_minutes),
      if (need_seconds) paste0(
        pillar::style_subtle(":"),
        pillar::style_num(format_two_digits(xx$second_of_minute), xx$sign, highlight_seconds)
      )
    )
    data <- paste0(
      data_seconds,
      if (need_seconds) {
        pillar::style_num(format_tics(xx$tics), xx$sign, highlight_tics)
      }
    )
  } else {
    data_seconds <- paste0(
      if (need_sign) ifelse(xx$sign, "-", " ") else "",
      pillar::style_num(format_two_digits(xx$minute_of_hour), xx$sign, highlight_minutes),
      pillar::style_subtle("'"),
      pillar::style_num(format_two_digits(xx$second_of_minute), xx$sign, highlight_seconds)
    )
    data <- paste0(
      data_seconds,
      pillar::style_num(format_tics(xx$tics), xx$sign, highlight_tics),
      pillar::style_subtle('"')
    )
  }

  na_indent <- crayon::col_nchar(data_seconds[1], type = "width") - 2L
  data[is.na(x)] <- NA

  pillar::new_pillar_shaft_simple(data, na_indent = na_indent)
}

# nocov start
.onLoad <- function(...) {
  vctrs::s3_register("pillar::pillar_shaft", "hms")

  if (utils::packageVersion("vctrs") <= "0.1.0") {
    vec_default_cast <<- vec_default_cast_old
  } else {
    rm("vec_default_cast", inherits = TRUE)
  }

  invisible()
}
# nocov end

format_hours <- function(x) {
  format(format_two_digits(x), justify = "right")
}

format_two_digits <- function(x) {
  formatC(x, format = "f", digits = 0, width = 2, flag = "0")
}

format_tics <- function(x) {
  x <- x / TICS_PER_SECOND

  out <- format(x, scientific = FALSE, digits = SPLIT_SECOND_DIGITS + 1L)
  digits <- max(min(max(nchar(out) - 2), SPLIT_SECOND_DIGITS), 0)
  out <- formatC(x, format = "f", digits = digits)
  gsub("^0", "", out)
}

SPLIT_SECOND_DIGITS <- 6L

TICS_PER_SECOND <- 10^SPLIT_SECOND_DIGITS
SECONDS_PER_MINUTE <- 60
MINUTES_PER_HOUR <- 60
HOURS_PER_DAY <- 24

TICS_PER_MINUTE <- SECONDS_PER_MINUTE * TICS_PER_SECOND
TICS_PER_HOUR <- MINUTES_PER_HOUR * TICS_PER_MINUTE
TICS_PER_DAY <- HOURS_PER_DAY * TICS_PER_HOUR

days <- function(x) {
  trunc(x / TICS_PER_DAY)
}

hours <- function(x) {
  trunc(x / TICS_PER_HOUR)
}

hour_of_day <- function(x) {
  abs(hours(x) - days(x) * HOURS_PER_DAY)
}

minutes <- function(x) {
  trunc(x / TICS_PER_MINUTE)
}

minute_of_hour <- function(x) {
  abs(minutes(x) - hours(x) * MINUTES_PER_HOUR)
}

seconds <- function(x) {
  trunc(x / TICS_PER_SECOND)
}

second_of_minute <- function(x) {
  abs(seconds(x) - minutes(x) * SECONDS_PER_MINUTE)
}

tics <- function(x) {
  x
}

tic_of_second <- function(x) {
  abs(tics(x) - seconds(x) * TICS_PER_SECOND)
}

decompose <- function(x) {
  x <- vec_data(x) * TICS_PER_SECOND

  # #140
  xr <- round(x)

  out <- list(
    sign = xr < 0 & !is.na(xr),
    hours = abs(hours(xr)),
    minute_of_hour = minute_of_hour(xr),
    second_of_minute = second_of_minute(xr),
    tics = tic_of_second(xr)
  )

  # #140: Make sure zeros are printed
  fake_zero <- (out$tics == 0) & (xr != x)
  out$tics[fake_zero] <- 0.25

  out
}

empty_tz <- function(x) {
  attr(x, "tzone") <- ""
  x
}

test_that("arithmetics work", {
  expect_equal(as.Date("2016-03-31") + hms(hours = 1), as.Date("2016-03-31"))
  expect_equal(as.Date("2016-03-31") + hms(days = -1), as.Date("2016-03-30"))
  expect_equal(empty_tz(as.POSIXct("2016-03-31") + hms(1)), as.POSIXct("2016-03-31 00:00:01"))
  expect_equal(hms(hours = 1) + as.Date("2016-03-31"), as.Date("2016-03-31"))
  expect_equal(hms(days = 1) + as.Date("2016-03-31"), as.Date("2016-04-01"))
  expect_equal(empty_tz(hms(hours = 1) + as.POSIXct("2016-03-31")), as.POSIXct("2016-03-31 01:00:00"))

  expect_difftime_equal(hms(1) + hms(2), hms(3))
  expect_difftime_equal(hms(1) - hms(2), hms(-1))
  expect_difftime_equal(2 * hms(1), hms(2))
  expect_difftime_equal(hms(hours = 1) / 2, hms(minutes = 30))
  expect_difftime_equal(-hms(1), hms(-1))
})

test_that("component extraction work", {
  x <- as.numeric(hms(12.3, 45, 23, 1)) * TICS_PER_SECOND
  expect_equal(tic_of_second(x), 300000)
  expect_equal(second_of_minute(x), 12)
  expect_equal(minute_of_hour(x), 45)
  expect_equal(hour_of_day(x), 23)
  expect_equal(days(x), 1)
})

test_that("component extraction work for negative times", {
  x <- as.numeric(-hms(12.3, 45, 23, 1)) * TICS_PER_SECOND
  expect_equal(tic_of_second(x), 300000)
  expect_equal(second_of_minute(x), 12)
  expect_equal(minute_of_hour(x), 45)
  expect_equal(hour_of_day(x), 23)
  expect_equal(days(x), -1)
})

#' Casting
#'
#' Double dispatch methods to support [vctrs::vec_cast()].
#'
#' @inheritParams vctrs::vec_cast
#'
#' @method vec_cast hms
#' @export
#' @export vec_cast.hms
vec_cast.hms <- function(x, to, ...) UseMethod("vec_cast.hms")

#' @method vec_cast.hms default
#' @export
vec_cast.hms.default <- function(x, to, ...) vec_default_cast(x, to, ...)

#' @method vec_cast.hms hms
#' @export
vec_cast.hms.hms <- function(x, to, ...) x

#' @method vec_cast.hms difftime
#' @export
vec_cast.hms.difftime <- function(x, to, ...) {
  units(x) <- "secs"
  new_hms(as.numeric(vec_data(x)))
}

#' @method vec_cast.difftime hms
#' @export
vec_cast.difftime.hms <- function(x, to, ...) {
  # as.difftime() doesn't change the class
  class(x) <- "difftime"
  vec_cast(x, to, ...)
}

#' @method vec_cast.hms POSIXct
#' @export
vec_cast.hms.POSIXct <- function(x, to, ...) {
  vec_cast(as.POSIXlt(x), to)
}

#' @method vec_cast.POSIXct hms
#' @export
vec_cast.POSIXct.hms <- function(x, to, ...) {
  structure(as.numeric(x), tzone = "UTC", class = c("POSIXct", "POSIXt"))
}

#' @method vec_cast.hms POSIXlt
#' @export
vec_cast.hms.POSIXlt <- function(x, to, ...) {
  hms(x$sec, x$min, x$hour)
}

#' @method vec_cast.POSIXlt hms
#' @export
vec_cast.POSIXlt.hms <- function(x, to, ...) {
  as.POSIXlt(vec_cast(x, new_datetime()))
}

#' @method vec_cast.hms double
#' @export
vec_cast.hms.double <- function(x, to, ...) new_hms(x)

#' @method vec_cast.double hms
#' @export
vec_cast.double.hms <- function(x, to, ...) vec_data(x)

#' @method vec_cast.hms integer
#' @export
vec_cast.hms.integer <- function(x, to, ...) new_hms(as.numeric(x))

#' @method vec_cast.integer hms
#' @export
vec_cast.integer.hms <- function(x, to, ...) as.integer(vec_data(x))

#' @method vec_cast.hms character
#' @export
vec_cast.hms.character <- function(x, to, ...) {
  ret <- parse_hms(x)
  lossy <- is.na(ret) & !is.na(x)
  abort_lossy_cast(x, to, ..., lossy = lossy)
  ret
}

#' @method vec_cast.character hms
#' @export
vec_cast.character.hms <- function(x, to, ...) format_hms(x)

# Requires vctrs > 0.1.0
vec_default_cast <- NULL

utils::globalVariables("vec_unspecified_cast")
vec_default_cast_old <- function(x, to, ...) {
  if (is.logical(x)) {
    vec_unspecified_cast(x, to, ...)
  } else {
    stop_incompatible_cast(x, to, ...)
  }
}

abort_lossy_cast <- function(x, to, ..., lossy) {
  problems <- which(lossy)
  if (is_empty(problems)) return()

  abort(
    paste0("Lossy cast from <character> to <hms> at position(s) ", commas(problems))
  )
}

commas <- function(problems) {
  MAX_BULLETS <- 6L
  if (length(problems) >= MAX_BULLETS) {
    n_more <- length(problems) - MAX_BULLETS + 1L
    problems[[MAX_BULLETS]] <-
      paste0(pre_dots("(and "), n_more, " more)")
    length(problems) <- MAX_BULLETS
  }

  paste0(problems, collapse = ", ")
}

pre_dots <- function(x) {
  if (length(x) > 0) {
    paste0("... ", x)
  } else {
    character()
  }
}

#' @details
#' `r lifecycle::badge("stable")`
#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom cli cli_abort
#' @importFrom cli cli_text
#' @importFrom cli cli_warn
#' @importFrom lifecycle deprecate_soft
#' @importFrom lifecycle deprecate_warn
#' @importFrom lifecycle expect_deprecated
#' @importFrom methods setOldClass
#' @importFrom pkgconfig get_config
#' @import rlang
#' @import vctrs
## usethis namespace: end
NULL

#' Round or truncate to a multiple of seconds
#'
#' Convenience functions to round or truncate to a multiple of seconds.
#' @param x A vector of class [hms]
#' @param secs Multiple of seconds, a positive numeric. Values less than one
#'   are supported
#' @param digits Number of digits, a whole number.
#'   Negative numbers are supported.
#' @return The input, rounded or truncated to the nearest multiple of `secs`
#'   (or number of `digits`)
#' @export
#' @examples
#' round_hms(as_hms("12:34:56"), 5)
#' round_hms(as_hms("12:34:56"), 60)
#' round_hms(as_hms("12:34:56.78"), 0.25)
#' round_hms(as_hms("12:34:56.78"), digits = 1)
#' round_hms(as_hms("12:34:56.78"), digits = -2)
round_hms <- function(x, secs = NULL, digits = NULL) {
  secs <- digits_to_secs(secs, digits)

  vec_restore(round(as.numeric(x) / secs) * secs, x)
}

#' @rdname round_hms
#' @export
#' @examples
#' trunc_hms(as_hms("12:34:56"), 60)
trunc_hms <- function(x, secs = NULL, digits = NULL) {
  secs <- digits_to_secs(secs, digits)

  vec_restore(trunc(as.numeric(x) / secs) * secs, x)
}

digits_to_secs <- function(secs, digits) {
  if (is.null(digits)) {
    if (is.null(secs)) {
      abort("Exactly one of `secs` or `digits` is required.")
    }

    return(secs)
  }

  if (!is.null(secs)) {
    abort("Exactly one of `secs` or `digits` is required.")
  }

  if (!is_integerish(digits)) {
    abort("`digits` must be a whole number")
  }

  if (digits >= -1) {
    secs <- 10^-digits
  } else if (digits == -2) {
    secs <- 60
  } else if (digits == -3) {
    secs <- 600
  } else if (digits <= -4) {
    secs <- 3600 * (10^(-digits - 4))
  }

  secs
}

test_that("round_hms()", {
  expect_equal(round_hms(parse_hms("12:34:56"), 5), hms(55, 34, 12))
  expect_equal(round_hms(parse_hms("12:34:56"), 60), hms(0, 35, 12))

  expect_equal(round_hms(parse_hms("12:34:56.78"), digits = 1), hms(56.8, 34, 12))
  expect_equal(round_hms(parse_hms("12:34:56.78"), digits = 0), hms(57, 34, 12))
  expect_equal(round_hms(parse_hms("12:34:36"), digits = -1), hms(40, 34, 12))

  expect_equal(round_hms(parse_hms("12:34:56"), digits = -2), hms(0, 35, 12))
  expect_equal(round_hms(parse_hms("12:34:56"), digits = -3), hms(0, 30, 12))
  expect_equal(round_hms(parse_hms("12:34:56"), digits = -4), hms(0, 0, 13))
  expect_equal(round_hms(parse_hms("12:34:56"), digits = -5), hms(0, 0, 10))
  expect_equal(round_hms(parse_hms("12:34:56"), digits = -6), hms(0, 0, 0))

  expect_equal(round_hms(hms(0.7), 0.25), hms(0.75))
  expect_equal(round_hms(hms(NA), 5), hms(NA))
  expect_equal(round_hms(parse_hms(c("12:34:56", NA)), 5), as_hms(c(hms(55, 34, 12), hms(NA))))
})

test_that("trunc_hms()", {
  expect_equal(trunc_hms(parse_hms("12:34:56"), 5), hms(55, 34, 12))
  expect_equal(trunc_hms(parse_hms("12:34:56"), 60), hms(0, 34, 12))

  expect_equal(trunc_hms(parse_hms("12:34:56.78"), digits = 1), hms(56.7, 34, 12))
  expect_equal(trunc_hms(parse_hms("12:34:56.78"), digits = 0), hms(56, 34, 12))
  expect_equal(trunc_hms(parse_hms("12:34:36"), digits = -1), hms(30, 34, 12))

  expect_equal(trunc_hms(parse_hms("12:34:56"), digits = -2), hms(0, 34, 12))
  expect_equal(trunc_hms(parse_hms("12:34:56"), digits = -3), hms(0, 30, 12))
  expect_equal(trunc_hms(parse_hms("12:34:56"), digits = -4), hms(0, 0, 12))
  expect_equal(trunc_hms(parse_hms("12:34:56"), digits = -5), hms(0, 0, 10))
  expect_equal(trunc_hms(parse_hms("12:34:56"), digits = -6), hms(0, 0, 0))

  expect_equal(trunc_hms(hms(0.7), 0.25), hms(0.5))
  expect_equal(trunc_hms(hms(NA), 5), hms(NA))
  expect_equal(trunc_hms(parse_hms(c("12:34:56", NA)), 5), as_hms(c(hms(55, 34, 12), hms(NA))))
})

#' Parsing hms values
#'
#' @description
#' These functions convert character vectors to objects of the [hms] class.
#' `NA` values are supported.
#'
#' `parse_hms()` accepts values of the form `"HH:MM:SS"`, with optional
#' fractional seconds.
#' @param x A character vector
#' @return An object of class [hms].
#'
#' @export
#' @examples
#' parse_hms("12:34:56")
#' parse_hms("12:34:56.789")
parse_hms <- function(x) {
  parse_time(x, format = "%H:%M:%OS")
}

#' @rdname parse_hms
#' @description
#' `parse_hm()` accepts values of the form `"HH:MM"`.
#' @export
#' @examples
#' parse_hm("12:34")
parse_hm <- function(x) {
  parse_time(x, format = "%H:%M")
}

parse_time <- function(x, format) {
  if (length(x) == 0) {
    return(hms())
  }

  parsed <- strptime(
    paste0("1970-01-01 ", as.character(x)),
    format = paste0("%Y-%m-%d ", format),
    tz = "UTC"
  )
  hms(
    seconds = parsed$sec,
    minutes = parsed$min,
    hours = parsed$hour + (parsed$mday - 1L) * 24L
  )
}

test_that("parse_hms", {
  expect_equal(parse_hms("12:34:56"), hms(56, 34, 12))
  expect_equal(parse_hms("12:34:56.789"), hms(56.789, 34, 12))
  expect_equal(parse_hms(NA), hms(NA))
  expect_equal(parse_hms(c("12:34:56", NA)), as_hms(c(hms(56, 34, 12), hms(NA))))
})

test_that("parse_hm", {
  expect_equal(parse_hm("12:34"), hms(0, 34, 12))
  expect_equal(parse_hm(NA), hms(NA))
  expect_equal(parse_hm(c("12:34", NA)), as_hms(c(hms(0, 34, 12), hms(NA))))
})

test_that("parse_hms", {
  expect_equal(parse_hms("12:34:56"), hms(56, 34, 12))
  expect_equal(parse_hms("12:34:56.789"), hms(56.789, 34, 12))
  expect_equal(parse_hms(NA), hms(NA))
  expect_equal(
    parse_hms(c("12:34:56", NA)),
    as_hms(c(hms(56, 34, 12), hms(NA)))
  )
  expect_identical(parse_hms("23:59:59.999999999"), as_hms(86399.999999999))
  expect_identical(parse_hms("24:00:00.999999999"), as_hms(86400.999999999))
  expect_identical(parse_hms(character()), hms())
})

test_that("parse_hm", {
  expect_identical(parse_hm("12:34"), hms(0, 34, 12))
  expect_identical(parse_hm(NA), hms(NA))
  expect_identical(parse_hm(c("12:34", NA)), as_hms(c(hms(0, 34, 12), hms(NA))))
  expect_identical(parse_hms(character()), hms())
})

#' @details
#' `r lifecycle::badge("stable")`
#' @importFrom lifecycle deprecate_soft expect_deprecated
#' @import vctrs
#' @import rlang
#' @aliases hms-package NULL
"_PACKAGE"

#' @importFrom methods setOldClass
setOldClass(c("hms", "difftime"))

#' A simple class for storing time-of-day values
#'
#' The values are stored as a [difftime] vector with a custom class,
#' and always with "seconds" as unit for robust coercion to numeric.
#' Supports construction from time values, coercion to and from
#' various data types, and formatting.  Can be used as a regular column in a
#' data frame.
#'
#' @name hms
#' @examples
#' hms(56, 34, 12)
#' hms()
#'
#' new_hms(as.numeric(1:3))
#' # Supports numeric only!
#' try(new_hms(1:3))
#'
#' as_hms(1)
#' as_hms("12:34:56")
#' as_hms(Sys.time())
#' as.POSIXct(hms(1))
#' data.frame(a = hms(1))
#' d <- data.frame(hours = 1:3)
#' d$hours <- hms(hours = d$hours)
#' d
NULL

# Construction ------------------------------------------------------------

#' hms()
#'
#' `hms()` is a high-level constructor that accepts second, minute, hour and day components
#' as numeric vectors.
#'
#' @rdname hms
#' @details For `hms()`, all arguments must have the same length or be
#'   `NULL`.  Odd combinations (e.g., passing only `seconds` and
#'   `hours` but not `minutes`) are rejected.
#' @param seconds,minutes,hours,days Time since midnight. No bounds checking is
#'   performed.
#' @export
hms <- function(seconds = NULL, minutes = NULL, hours = NULL, days = NULL) {
  args <- list(seconds = seconds, minutes = minutes, hours = hours, days = days)
  check_args(args)
  arg_secs <- map2(args, c(1, 60, 3600, 86400), `*`)
  secs <- reduce(arg_secs[!map_lgl(args, is.null)], `+`)
  if (is.null(secs)) secs <- numeric()

  new_hms(as.numeric(secs))
}

#' new_hms()
#'
#' `new_hms()` is a low-level constructor that only checks that its input has the correct base type, [numeric].
#'
#' @rdname hms
#' @export
new_hms <- function(x = numeric()) {
  vec_assert(x, numeric())

  out <- new_duration(x, units = "secs")

  # no class argument?
  class(out) <- c("hms", class(out))
  out
}

#' is_hms()
#'
#' `is_hms()` checks if an object is of class `hms`.
#'
#' @rdname hms
#' @export
is_hms <- function(x) inherits(x, "hms")

#' Deprecated functions
#'
#' @name Deprecated
NULL

#' Deprecated is.hms()
#'
#' `is.hms()` has been replaced by [is_hms()].
#'
#' @inheritParams is_hms
#' @rdname Deprecated
#' @export
#' @keywords internal
is.hms <- function(x) {
  deprecate_soft("0.5.0", "hms::is.hms()", "hms::is_hms()")
  is_hms(x)
}

#' @export
vec_ptype_abbr.hms <- function(x, ...) {
  "time"
}

#' @export
vec_ptype_full.hms <- function(x, ...) {
  "time"
}

# Coercion in -------------------------------------------------------------

#' as_hms()
#'
#' `as_hms()` is a generic that supports conversions beyond casting.
#' The default method forwards to [vec_cast()].
#'
#' For arguments of type [POSIXct] and [POSIXlt], `as_hms()` does not perform timezone
#' conversion.
#' Use [lubridate::with_tz()] and [lubridate::force_tz()] as necessary.
#'
#' @rdname hms
#' @param x An object.
#' @export
as_hms <- function(x, ...) {
  check_dots_used()

  UseMethod("as_hms")
}

#' @export
as_hms.default <- function(x, ...) {
  vec_cast(x, new_hms())
}

#' Deprecated as.hms()
#'
#' `as.hms()` has been replaced by [as_hms()], which does not have a `tz` argument.
#' Change the timezone before converting if necessary, e.g. using [lubridate::with_tz()].
#'
#' @inheritParams as_hms
#' @param ... Arguments passed on to further methods.
#' @rdname Deprecated
#' @export
#' @keywords internal
as.hms <- function(x, ...) {
  deprecate_soft("0.5.0", "hms::as.hms()", "hms::as_hms()")
  UseMethod("as.hms", x)
}

#' @rdname Deprecated
#' @export
as.hms.default <- function(x, ...) {
  as_hms(x)
}

#' @rdname Deprecated
#' @param tz The time zone in which to interpret a POSIXt time for extracting
#'   the time of day.  The default is now the zone of `x` but was `"UTC"`
#'   for v0.3 and earlier.  The previous behavior can be restored by calling
#'   `pkgconfig::set_config("hms::default_tz", "UTC")`, see
#'   [pkgconfig::set_config()].
#' @export
#' @importFrom pkgconfig get_config
as.hms.POSIXt <- function(x, tz = pkgconfig::get_config("hms::default_tz", ""), ...) {
  time <- as.POSIXlt(x, tz = tz)
  vec_cast(time, new_hms())
}

#' @rdname Deprecated
#' @export
as.hms.POSIXlt <- function(x, tz = pkgconfig::get_config("hms::default_tz", ""), ...) {
  # We need to roundtrip via as.POSIXct() to respect the time zone
  time <- as.POSIXlt(as.POSIXct(x), tz = tz)
  vec_cast(time, new_hms())
}


# Coercion out ------------------------------------------------------------

#' @rdname hms
#' @inheritParams base::as.data.frame
#' @export
as.POSIXct.hms <- function(x, ...) {
  vec_cast(x, new_datetime())
}

#' @rdname hms
#' @export
as.POSIXlt.hms <- function(x, ...) {
  vec_cast(x, as.POSIXlt(new_datetime()))
}

#' @rdname hms
#' @export
as.character.hms <- function(x, ...) {
  vec_cast(x, character())
}

format_hms <- function(x) {
  xx <- decompose(x)

  ifelse(is.na(x), NA_character_, paste0(
    ifelse(xx$sign, "-", ""),
    format_hours(xx$hours), ":",
    format_two_digits(xx$minute_of_hour), ":",
    format_two_digits(xx$second_of_minute),
    format_tics(xx$tics)))
}


# Subsetting --------------------------------------------------------------

#' @export
`[[.hms` <- function(x, ...) {
  vec_restore(NextMethod(), x)
}

#' @export
`[<-.hms` <- function(x, i, value) {
  if (missing(i)) {
    i <- TRUE
  }

  x <- vec_data(x)

  # Workaround for Ops.difftime() implementation for unary minus
  if (identical(class(value), "numeric")) {
    attr(value, "units") <- NULL
  }

  value <- vec_cast(value, new_hms())
  x[i] <- vec_data(value)
  new_hms(x)
}

# Combination -------------------------------------------------------------
#' @export
c.hms <- function(x, ...) {
  # Needed to override c.difftime()
  vec_c(x, ...)
}

# Updating ----------------------------------------------------------------

#' @export
`units<-.hms` <- function(x, value) {
  if (!identical(value, "secs")) {
    warning("hms always uses seconds as unit.", call. = FALSE)
  }
  x
}


# Output ------------------------------------------------------------------

#' @rdname hms
#' @export
format.hms <- function(x, ...) {
  if (length(x) == 0L) {
    "hms()"
  } else {
    format(as.character(x), justify = "right")
  }
}

#' @rdname hms
#' @export
print.hms <- function(x, ...) {
  cat(format(x), sep = "\n")
  invisible(x)
}

test_that("unique", {
  expect_identical(unique(new_hms(numeric(0))), new_hms(numeric(0)))
  expect_identical(unique(new_hms(NA_real_)), new_hms(NA_real_))
  expect_identical(unique(new_hms(101)), new_hms(101))
  expect_identical(unique(new_hms(c(101, 101, 99, 101))), new_hms(c(101, 99)))
  expect_identical(unique(new_hms(c(101, NA_real_, 101, 99, 101))), new_hms(c(101, NA_real_, 99)))
})

test_that("Can't update units", {
  x <- hms(minutes = 3)
  expect_equal(units(x), "secs")
  expect_warning(units(x) <- "mins", "always uses seconds")
  expect_equal(units(x), "secs")
  expect_warning(units(x) <- "secs", NA)
  expect_equal(units(x), "secs")
})

test_that("coercion in", {
  expect_identical(as_hms(0.5 * 86400), hms(hours = 12))
  expect_identical(as_hms(-0.25 * 86400), hms(hours = -6))
  expect_hms_equal(as_hms("12:34:56"), hms(56, 34, 12))
  expect_hms_equal(as_hms(strptime("12:34:56", format = "%H:%M:%S", tz = "UTC")),
    hms(56, 34, 12))
  expect_hms_equal(as_hms(strptime("12:34:56", format = "%H:%M:%S", tz = "Europe/Zurich")),
    hms(56, 34, 12))
  expect_hms_equal(as_hms(strptime("12:34:56", format = "%H:%M:%S", tz = "PST8PDT")),
    hms(56, 34, 12))
  expect_hms_equal(as_hms(as.POSIXct(strptime("12:34:56", format = "%H:%M:%S", tz = "UTC"))),
    hms(56, 34, 12))
  expect_hms_equal(as_hms(as.POSIXct(strptime("12:34:56", format = "%H:%M:%S", tz = "Europe/Zurich"))),
    hms(56, 34, 12))
  expect_hms_equal(as_hms(as.POSIXct(strptime("12:34:56", format = "%H:%M:%S", tz = "PST8PDT"))),
    hms(56, 34, 12))

  now <- Sys.time()
  now_lt <- as.POSIXlt(now)
  expect_hms_equal(as_hms(now), hms(now_lt$sec, now_lt$min, now_lt$hour))
  expect_hms_equal(as_hms(now_lt), as_hms(now))

  expect_error(as_hms(FALSE))

  x <- c("12:34:56", "ab:cd:ef")
  expect_error(hms::as_hms(x))
})

test_that("coercion out", {
  expect_identical(as.character(hms(56, 34, 12)), "12:34:56")
  expect_identical(as.character(hms(NA)), NA_character_)
  expect_identical(as.POSIXlt(hms(hours = 6)),
    strptime("1970-01-01 06:00:00",
      format = "%Y-%m-%d %H:%M:%S", tz = "UTC"))
  expect_identical(as.POSIXct(hms(hours = -6)),
    strptime("1970-01-01 18:00:00",
      format = "%Y-%m-%d %H:%M:%S", tz = "UTC") - 86400)

  df <- data.frame(a = 1:3)
  df$b <- hms(hours = df$a)
  expect_identical(df, data.frame(a = 1:3, b = hms(hours = 1:3)))
})

test_that("coercion in", {
  expect_identical(as_hms(0.5 * 86400), hms(hours = 12))
  expect_identical(as_hms(-0.25 * 86400), hms(hours = -6))
  expect_hms_equal(as_hms("12:34:56"), hms(56, 34, 12))
  expect_hms_equal(
    as_hms(strptime("12:34:56", format = "%H:%M:%S", tz = "UTC")),
    hms(56, 34, 12)
  )
  expect_hms_equal(
    as_hms(strptime("12:34:56", format = "%H:%M:%S", tz = "Europe/Zurich")),
    hms(56, 34, 12)
  )
  expect_hms_equal(
    as_hms(strptime("12:34:56", format = "%H:%M:%S", tz = "PST8PDT")),
    hms(56, 34, 12)
  )
  expect_hms_equal(
    as_hms(as.POSIXct(strptime("12:34:56", format = "%H:%M:%S", tz = "UTC"))),
    hms(56, 34, 12)
  )
  expect_hms_equal(
    as_hms(as.POSIXct(strptime(
      "12:34:56",
      format = "%H:%M:%S",
      tz = "Europe/Zurich"
    ))),
    hms(56, 34, 12)
  )
  expect_hms_equal(
    as_hms(as.POSIXct(strptime(
      "12:34:56",
      format = "%H:%M:%S",
      tz = "PST8PDT"
    ))),
    hms(56, 34, 12)
  )

  now <- Sys.time()
  now_lt <- as.POSIXlt(now)
  expect_hms_equal(as_hms(now), hms(now_lt$sec, now_lt$min, now_lt$hour))
  expect_hms_equal(as_hms(now_lt), as_hms(now))

  expect_error(as_hms(FALSE))

  x <- c("12:34:56", "ab:cd:ef")
  expect_error(hms::as_hms(x))
})

test_that("coercion out", {
  expect_identical(as.character(hms(56, 34, 12)), "12:34:56")
  expect_identical(as.character(hms(NA)), NA_character_)
  expect_identical(
    as.POSIXlt(hms(hours = 6)),
    strptime("1970-01-01 06:00:00", format = "%Y-%m-%d %H:%M:%S", tz = "UTC")
  )
  expect_identical(
    as.POSIXct(hms(hours = -6)),
    strptime("1970-01-01 18:00:00", format = "%Y-%m-%d %H:%M:%S", tz = "UTC") -
      86400
  )

  df <- data.frame(a = 1:3)
  df$b <- hms(hours = df$a)
  expect_identical(df, data.frame(a = 1:3, b = hms(hours = 1:3)))
})

expect_hms_equal <- function(x, y) {
  expect_s3_class(x, "hms")
  expect_s3_class(y, "hms")
  expect_equal(as.numeric(x), as.numeric(y))
}

expect_difftime_equal <- function(x, y) {
  expect_s3_class(x, "difftime")
  expect_s3_class(y, "difftime")
  expect_equal(as.numeric(as_hms(x)), as.numeric(as_hms(y)))
}

test_that("constructor", {
  expect_identical(hms(1:3, 2:4, 3:5, 4:6),
    hms(seconds = 1:3 + 2:4 * 60 + 3:5 * 3600 + 4:6 * 86400))
  expect_identical(hms(-1, 1), hms(59))
  expect_identical(hms(3600), hms(hours = 1))

  expect_equal(length(hms(1)), 1L)
  expect_true(is_hms(hms(1)))
  expect_s3_class(hms(1), "difftime")
  expect_identical(as.numeric(hms(1)), 1)
  expect_identical(as.difftime(hms(1)), hms(1))
})


test_that("casting", {
  expect_identical(units(as_hms(as.difftime(1, units = "mins"))), "secs")
  expect_identical(as_hms(hms(1)), hms(1))

  expect_identical(as_hms(as.difftime(1:3, units = "secs")), hms(as.numeric(1:3)))
})

test_that("zero length (#35)", {
  expect_equal(length(hms()), 0L)
  expect_true(is_hms(hms()))
  expect_s3_class(hms(), "difftime")
  expect_identical(as.numeric(hms()), numeric())
  expect_identical(as.difftime(hms()), hms())

  expect_identical(hms(), hms(seconds = numeric()))
  expect_identical(hms(), hms(minutes = numeric()))
  expect_identical(hms(), hms(hours = numeric()))
  expect_identical(hms(), hms(days = numeric()))
  expect_identical(hms(), as_hms(numeric()))
})

test_that("bad input", {
  expect_error(hms(hours = 1, seconds = 3), "only")
  expect_error(hms(minutes = 1, days = 3), "only")
  expect_error(hms(minutes = 1, hours = 2:3), "same length or be NULL")
  expect_error(hms(seconds = 1:5, minutes = 6:10, hours = 11:17), "same length or be NULL")
  expect_error(hms("05:00"), "must be numeric")
})

test_that("is.hms()", {
  expect_deprecated(expect_identical(is.hms(hms), is_hms(hms)))
  expect_deprecated(expect_identical(is.hms(3), is_hms(3)))
})

test_that("range subsetting keeps class", {
  expect_identical(hms(1:3)[2], hms(2))
  expect_identical(hms(1:3)[2:3], hms(2:3))
})

test_that("range updating keeps class", {
  skip_if(packageVersion("vctrs") <= "0.1.0")

  x <- hms(1:3)
  x[2] <- hms(4)
  expect_identical(x, hms(c(1, 4, 3)))
  x <- hms(1:4)
  x[2:3] <- hms(5:6)
  expect_identical(x, hms(c(1, 5, 6, 4)))
})

test_that("range updating warns if lossy cast", {
  skip_if(packageVersion("vctrs") <= "0.1.0")

  x <- hms(1:3)

  expect_error(x[2] <- "a")
})

test_that("index subsetting keeps class", {
  expect_identical(hms(1:3)[[2]], hms(2))
})

test_that("index updating keeps class", {
  x <- hms(1:3)
  x[[2]] <- 4
  expect_identical(x, hms(c(1, 4, 3)))
})

test_that("coercion in", {
  expect_deprecated(
    expect_identical(as.hms(0.5 * 86400), hms(hours = 12))
  )
  expect_deprecated(
    expect_identical(as.hms(-0.25 * 86400), hms(hours = -6))
  )
  expect_deprecated(
    expect_hms_equal(as.hms("12:34:56"), hms(56, 34, 12))
  )
  expect_deprecated(
    expect_hms_equal(
      as.hms(strptime("12:34:56", format = "%H:%M:%S", tz = "UTC"), tz = "UTC"),
      hms(56, 34, 12)
    )
  )
  expect_deprecated(
    expect_hms_equal(
      as.hms(strptime("12:34:56", format = "%H:%M:%S", tz = "Europe/Zurich"), tz = "Europe/Zurich"),
      hms(56, 34, 12)
    )
  )
  expect_deprecated(
    expect_hms_equal(
      as.hms(strptime("12:34:56", format = "%H:%M:%S", tz = "PST8PDT"), tz = "PST8PDT"),
      hms(56, 34, 12)
    )
  )

  now <- Sys.time()
  now_lt <- as.POSIXlt(now)
  expect_deprecated(
    expect_hms_equal(as.hms(now), hms(now_lt$sec, now_lt$min, now_lt$hour))
  )
  expect_deprecated(
    expect_hms_equal(as.hms(now_lt), as_hms(now))
  )

  expect_error(expect_deprecated(as.hms(FALSE)))
})

test_that("coercion out", {
  expect_identical(as.character(hms(56, 34, 12)), "12:34:56")
  expect_identical(as.character(hms(NA)), NA_character_)
  expect_identical(as.POSIXlt(hms(hours = 6)),
    strptime("1970-01-01 06:00:00",
      format = "%Y-%m-%d %H:%M:%S", tz = "UTC"))
  expect_identical(as.POSIXct(hms(hours = -6)),
    strptime("1970-01-01 18:00:00",
      format = "%Y-%m-%d %H:%M:%S", tz = "UTC") - 86400)

  df <- data.frame(a = 1:3)
  df$b <- hms(hours = df$a)
  expect_identical(df, data.frame(a = 1:3, b = hms(hours = 1:3)))
})

test_that("combination keeps class and order", {
  expect_identical(c(hms(1), hms(2)), hms(1:2))
})

test_that("combination errs if not supported", {
  expect_error(c(hms(1), factor(2)))
})

test_that("combination coerces to hms", {
  expect_identical(c(hms(1), hms(2)), hms(1:2))
  if (getRversion() < "3.3") skip("Only for R >= 3.3")

  expect_identical(c(hms(1), new_duration(2)), hms(1:2))
  expect_identical(vec_c(new_duration(1), hms(2)), hms(1:2))
})


# In R base,`c(as.difftime("20:00:00"), NA)` fails
test_that("composition with NA works", {
  expect_identical(
    c(hms(1), NA),
    hms(c(1, NA))
  )

  expect_identical(
    vec_c(hms(1), NA),
    hms(vec_c(1, NA))
  )

  h <- hms(1)
  expect_equal(vec_ptype2(h, NA), vec_ptype(h))
  expect_equal(vec_ptype2(NA, h), vec_ptype(h))
})

test_that("combination keeps class and order", {
  expect_identical(c(hms(1), hms(2)), hms(1:2))
})

test_that("combination errs if not supported", {
  expect_error(c(hms(1), factor(2)))
})

test_that("combination coerces to hms", {
  expect_identical(c(hms(1), hms(2)), hms(1:2))
  if (getRversion() < "3.3") {
    skip("Only for R >= 3.3")
  }

  expect_identical(c(hms(1), new_duration(2)), hms(1:2))
  expect_identical(vec_c(new_duration(1), hms(2)), hms(1:2))
})


# In R base,`c(as.difftime("20:00:00"), NA)` fails
test_that("composition with NA works", {
  expect_identical(
    c(hms(1), NA),
    hms(c(1, NA))
  )

  expect_identical(
    vec_c(hms(1), NA),
    hms(vec_c(1, NA))
  )

  h <- hms(1)
  expect_equal(vec_ptype2(h, NA), vec_ptype(h))
  expect_equal(vec_ptype2(NA, h), vec_ptype(h))
})

test_that("duration", {
  skip_if_not_installed("lubridate")
  expect_identical(lubridate::as.duration(hms(minutes = 1:3)),
    lubridate::duration(minutes = 1:3))
})

test_that("interval", {
  skip_if_not_installed("lubridate")
  timestamp <- Sys.time()
  expect_identical(lubridate::as.interval(hms(seconds = 2), timestamp),
    lubridate::interval(timestamp, timestamp + 2))
})

test_that("period", {
  skip_if_not_installed("lubridate")
  expect_identical(lubridate::as.period(hms(hours = -1)),
    lubridate::period(hours = -1))
})

test_that("duration", {
  skip_if_not_installed("lubridate")
  expect_identical(
    lubridate::as.duration(hms(minutes = 1:3)),
    lubridate::duration(minutes = 1:3)
  )
})

test_that("interval", {
  skip_if_not_installed("lubridate")
  timestamp <- Sys.time()
  expect_identical(
    lubridate::as.interval(hms(seconds = 2), timestamp),
    lubridate::interval(timestamp, timestamp + 2)
  )
})

test_that("period", {
  skip_if_not_installed("lubridate")
  expect_identical(
    lubridate::as.period(hms(hours = -1)),
    lubridate::period(hours = -1)
  )
})

test_that("seq.hms works", {
  expect_snapshot(
    error = TRUE,
    {
      seq(hms(1), 1)
      seq(hms(4), hms(1), by = hms(1))
      seq(hms(1), hms(4), by = hms(-1))
      seq(hms(1), hms(4), by = "test")
      seq(hms(1), hms(4), by = hms())
      seq(hms(1), hms(4), by = hms(0))
    }
  )

  expect_snapshot({
    seq(hms(1), hms(4), along.with = 1:2)
    seq(hms(1), hms(4), length.out = 2)

    seq(hms(1), hms(4))
    seq(hms(1), hms(4), by = hms(1))

    seq(hms(4), hms(1))
    seq(hms(4), hms(1), by = hms(-1))
  })
})

check_args <- function(args, call = caller_env()) {
  is_null <- map_lgl(args, is.null)
  if (all(is_null)) {
    return()
  }

  valid <- map_lgl(args[!is_null], is_numeric_or_na)
  if (!all(valid)) {
    cli_abort(
      "All arguments must be {.cls numeric} or {.val {NA}}",
      call = call
    )
  }

  if (!all(diff(which(!is_null)) == 1L)) {
    cli_abort(
      paste(
        "Can't pass only {names(is_null)[!is_null]} to {.fn hms}."
      ),
      call = call
    )
  }
}

is_numeric_or_na <- function(x) {
  is.numeric(x) || all(is.na(x))
}

# nocov start - compat-purrr (last updated: rlang 0.1.9000)

# This file serves as a reference for compatibility functions for
# purrr. They are not drop-in replacements but allow a similar style
# of programming. This is useful in cases where purrr is too heavy a
# package to depend on. Please find the most recent version in rlang's
# repository.

map <- function(.x, .f, ...) {
  lapply(.x, .f, ...)
}
map_mold <- function(.x, .f, .mold, ...) {
  out <- vapply(.x, .f, .mold, ..., USE.NAMES = FALSE)
  names(out) <- names(.x)
  out
}
map_lgl <- function(.x, .f, ...) {
  map_mold(.x, .f, logical(1), ...)
}
map_int <- function(.x, .f, ...) {
  map_mold(.x, .f, integer(1), ...)
}
map_dbl <- function(.x, .f, ...) {
  map_mold(.x, .f, double(1), ...)
}
map_chr <- function(.x, .f, ...) {
  map_mold(.x, .f, character(1), ...)
}
map_cpl <- function(.x, .f, ...) {
  map_mold(.x, .f, complex(1), ...)
}

pluck <- function(.x, .f) {
  map(.x, `[[`, .f)
}
pluck_lgl <- function(.x, .f) {
  map_lgl(.x, `[[`, .f)
}
pluck_int <- function(.x, .f) {
  map_int(.x, `[[`, .f)
}
pluck_dbl <- function(.x, .f) {
  map_dbl(.x, `[[`, .f)
}
pluck_chr <- function(.x, .f) {
  map_chr(.x, `[[`, .f)
}
pluck_cpl <- function(.x, .f) {
  map_cpl(.x, `[[`, .f)
}

map2 <- function(.x, .y, .f, ...) {
  Map(.f, .x, .y, ...)
}
map2_lgl <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "logical")
}
map2_int <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "integer")
}
map2_dbl <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "double")
}
map2_chr <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "character")
}
map2_cpl <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "complex")
}

args_recycle <- function(args) {
  lengths <- map_int(args, length)
  n <- max(lengths)

  stopifnot(all(lengths == 1L | lengths == n))
  to_recycle <- lengths == 1L
  args[to_recycle] <- map(args[to_recycle], function(x) rep.int(x, n))

  args
}
pmap <- function(.l, .f, ...) {
  args <- args_recycle(.l)
  do.call("mapply", c(
    FUN = list(quote(.f)),
    args, MoreArgs = quote(list(...)),
    SIMPLIFY = FALSE, USE.NAMES = FALSE
  ))
}

probe <- function(.x, .p, ...) {
  if (is_logical(.p)) {
    stopifnot(length(.p) == length(.x))
    .p
  } else {
    map_lgl(.x, .p, ...)
  }
}

keep <- function(.x, .f, ...) {
  .x[probe(.x, .f, ...)]
}
discard <- function(.x, .p, ...) {
  sel <- probe(.x, .p, ...)
  .x[is.na(sel) | !sel]
}
map_if <- function(.x, .p, .f, ...) {
  matches <- probe(.x, .p)
  .x[matches] <- map(.x[matches], .f, ...)
  .x
}

compact <- function(.x) {
  Filter(length, .x)
}

transpose <- function(.l) {
  inner_names <- names(.l[[1]])
  if (is.null(inner_names)) {
    fields <- seq_along(.l[[1]])
  } else {
    fields <- set_names(inner_names)
  }

  map(fields, function(i) {
    map(.l, .subset2, i)
  })
}

every <- function(.x, .p, ...) {
  for (i in seq_along(.x)) {
    if (!rlang::is_true(.p(.x[[i]], ...))) return(FALSE)
  }
  TRUE
}
some <- function(.x, .p, ...) {
  for (i in seq_along(.x)) {
    if (rlang::is_true(.p(.x[[i]], ...))) return(TRUE)
  }
  FALSE
}
negate <- function(.p) {
  function(...) !.p(...)
}

reduce <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(x, y, ...)
  Reduce(f, .x, init = .init)
}
reduce_right <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(y, x, ...)
  Reduce(f, .x, init = .init, right = TRUE)
}
accumulate <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(x, y, ...)
  Reduce(f, .x, init = .init, accumulate = TRUE)
}
accumulate_right <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(y, x, ...)
  Reduce(f, .x, init = .init, right = TRUE, accumulate = TRUE)
}

# nocov end

#' Coercion
#'
#' Double dispatch methods to support [vctrs::vec_ptype2()].
#'
#' @inheritParams vctrs::vec_ptype2
#'
#' @method vec_ptype2 hms
#' @export
#' @export vec_ptype2.hms
vec_ptype2.hms <- function(x, y, ..., x_arg = "", y_arg = "") UseMethod("vec_ptype2.hms", y)

#' @method vec_ptype2.hms default
#' @export
vec_ptype2.hms.default <- function(x, y, ..., x_arg = "", y_arg = "") {
  vec_default_ptype2(x, y, x_arg = x_arg, y_arg = y_arg)
}

#' @method vec_ptype2.hms hms
#' @export
vec_ptype2.hms.hms <- function(x, y, ...) hms::hms()

#' @method vec_ptype2.difftime hms
#' @export
vec_ptype2.difftime.hms <- function(x, y, ...) new_hms()

#' @method vec_ptype2.hms difftime
#' @export
vec_ptype2.hms.difftime <- function(x, y, ...) new_hms()

# seq.hms works

    Code
      seq(hms(1), 1)
    Condition
      Error in `seq()`:
      ! `to` isn't of class `hms` (current class: `numeric`).
    Code
      seq(hms(4), hms(1), by = hms(1))
    Condition
      Error in `seq.default()`:
      ! wrong sign in 'by' argument
    Code
      seq(hms(1), hms(4), by = hms(-1))
    Condition
      Error in `seq.default()`:
      ! wrong sign in 'by' argument
    Code
      seq(hms(1), hms(4), by = "test")
    Condition
      Error in `seq()`:
      ! `by` isn't of class `hms` or `difftime` (current class: `character`).
    Code
      seq(hms(1), hms(4), by = hms())
    Condition
      Error in `seq.default()`:
      ! 'by' must be of length 1
    Code
      seq(hms(1), hms(4), by = hms(0))
    Condition
      Error in `seq.default()`:
      ! invalid '(to - from)/by'

---

    Code
      seq(hms(1), hms(4), along.with = 1:2)
    Output
      00:00:01
      00:00:04
    Code
      seq(hms(1), hms(4), length.out = 2)
    Output
      00:00:01
      00:00:04
    Code
      seq(hms(1), hms(4))
    Output
      00:00:01
      00:00:02
      00:00:03
      00:00:04
    Code
      seq(hms(1), hms(4), by = hms(1))
    Output
      00:00:01
      00:00:02
      00:00:03
      00:00:04
    Code
      seq(hms(4), hms(1))
    Output
      00:00:04
      00:00:03
      00:00:02
      00:00:01
    Code
      seq(hms(4), hms(1), by = hms(-1))
    Output
      00:00:04
      00:00:03
      00:00:02
      00:00:01

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cast.R
\name{vec_cast.hms}
\alias{vec_cast.hms}
\title{Casting}
\usage{
\method{vec_cast}{hms}(x, to, ...)
}
\arguments{
\item{x}{Vectors to cast.}

\item{to}{Type to cast to. If \code{NULL}, \code{x} will be returned as is.}

\item{...}{For \code{vec_cast_common()}, vectors to cast. For
\code{vec_cast()}, \code{vec_cast_default()}, and \code{vec_restore()}, these
dots are only for future extensions and should be empty.}
}
\description{
Double dispatch methods to support \code{\link[vctrs:vec_cast]{vctrs::vec_cast()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/round.R
\name{round_hms}
\alias{round_hms}
\alias{trunc_hms}
\title{Round or truncate to a multiple of seconds}
\usage{
round_hms(x, secs = NULL, digits = NULL)

trunc_hms(x, secs = NULL, digits = NULL)
}
\arguments{
\item{x}{A vector of class \link{hms}}

\item{secs}{Multiple of seconds, a positive numeric. Values less than one
are supported}

\item{digits}{Number of digits, a whole number.
Negative numbers are supported.}
}
\value{
The input, rounded or truncated to the nearest multiple of \code{secs}
(or number of \code{digits})
}
\description{
Convenience functions to round or truncate to a multiple of seconds.
}
\examples{
round_hms(as_hms("12:34:56"), 5)
round_hms(as_hms("12:34:56"), 60)
round_hms(as_hms("12:34:56.78"), 0.25)
round_hms(as_hms("12:34:56.78"), digits = 1)
round_hms(as_hms("12:34:56.78"), digits = -2)
trunc_hms(as_hms("12:34:56"), 60)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse.R
\name{parse_hms}
\alias{parse_hms}
\alias{parse_hm}
\title{Parsing hms values}
\usage{
parse_hms(x)

parse_hm(x)
}
\arguments{
\item{x}{A character vector}
}
\value{
An object of class \link{hms}.
}
\description{
These functions convert character vectors to objects of the \link{hms} class.
\code{NA} values are supported.

\code{parse_hms()} accepts values of the form \code{"HH:MM:SS"}, with optional
fractional seconds.

\code{parse_hm()} accepts values of the form \code{"HH:MM"}.
}
\examples{
parse_hms("12:34:56")
parse_hms("12:34:56.789")
parse_hm("12:34")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{vec_ptype2.hms}
\alias{vec_ptype2.hms}
\title{Coercion}
\usage{
\method{vec_ptype2}{hms}(x, y, ..., x_arg = "", y_arg = "")
}
\arguments{
\item{x, y}{Vector types.}

\item{...}{These dots are for future extensions and must be empty.}

\item{x_arg, y_arg}{Argument names for \code{x} and \code{y}. These are used
in error messages to inform the user about the locations of
incompatible types (see \code{\link[vctrs:stop_incompatible_type]{stop_incompatible_type()}}).}
}
\description{
Double dispatch methods to support \code{\link[vctrs:vec_ptype2]{vctrs::vec_ptype2()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hms.R
\name{hms}
\alias{hms}
\alias{new_hms}
\alias{is_hms}
\alias{as_hms}
\alias{as.POSIXct.hms}
\alias{as.POSIXlt.hms}
\alias{as.character.hms}
\alias{format.hms}
\alias{print.hms}
\title{A simple class for storing time-of-day values}
\usage{
hms(seconds = NULL, minutes = NULL, hours = NULL, days = NULL)

new_hms(x = numeric())

is_hms(x)

as_hms(x, ...)

\method{as.POSIXct}{hms}(x, ...)

\method{as.POSIXlt}{hms}(x, ...)

\method{as.character}{hms}(x, ...)

\method{format}{hms}(x, ...)

\method{print}{hms}(x, ...)
}
\arguments{
\item{seconds, minutes, hours, days}{Time since midnight. No bounds checking is
performed.}

\item{x}{An object.}

\item{...}{additional arguments to be passed to or from methods.}
}
\description{
The values are stored as a \link{difftime} vector with a custom class,
and always with "seconds" as unit for robust coercion to numeric.
Supports construction from time values, coercion to and from
various data types, and formatting.  Can be used as a regular column in a
data frame.

\code{hms()} is a high-level constructor that accepts second, minute, hour and day components
as numeric vectors.

\code{new_hms()} is a low-level constructor that only checks that its input has the correct base type, \link{numeric}.

\code{is_hms()} checks if an object is of class \code{hms}.

\code{as_hms()} is a generic that supports conversions beyond casting.
The default method forwards to \code{\link[=vec_cast]{vec_cast()}}.
}
\details{
For \code{hms()}, all arguments must have the same length or be
\code{NULL}.  Odd combinations (e.g., passing only \code{seconds} and
\code{hours} but not \code{minutes}) are rejected.

For arguments of type \link{POSIXct} and \link{POSIXlt}, \code{as_hms()} does not perform timezone
conversion.
Use \code{\link[lubridate:with_tz]{lubridate::with_tz()}} and \code{\link[lubridate:force_tz]{lubridate::force_tz()}} as necessary.
}
\examples{
hms(56, 34, 12)
hms()

new_hms(as.numeric(1:3))
# Supports numeric only!
try(new_hms(1:3))

as_hms(1)
as_hms("12:34:56")
as_hms(Sys.time())
as.POSIXct(hms(1))
data.frame(a = hms(1))
d <- data.frame(hours = 1:3)
d$hours <- hms(hours = d$hours)
d
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hms.R
\docType{package}
\name{hms-package}
\alias{hms-package}
\title{hms: Pretty Time of Day}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Implements an S3 class for storing and formatting time-of-day values, based on the 'difftime' class.
}
\details{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://hms.tidyverse.org/}
  \item \url{https://github.com/tidyverse/hms}
  \item Report bugs at \url{https://github.com/tidyverse/hms/issues}
}

}
\author{
\strong{Maintainer}: Kirill Müller \email{kirill@cynkra.com} (\href{https://orcid.org/0000-0002-1416-3412}{ORCID})

Other contributors:
\itemize{
  \item R Consortium [funder]
  \item RStudio [funder]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hms.R
\name{Deprecated}
\alias{Deprecated}
\alias{is.hms}
\alias{as.hms}
\alias{as.hms.default}
\alias{as.hms.POSIXt}
\alias{as.hms.POSIXlt}
\title{Deprecated functions}
\usage{
is.hms(x)

as.hms(x, ...)

\method{as.hms}{default}(x, ...)

\method{as.hms}{POSIXt}(x, tz = pkgconfig::get_config("hms::default_tz", ""), ...)

\method{as.hms}{POSIXlt}(x, tz = pkgconfig::get_config("hms::default_tz", ""), ...)
}
\arguments{
\item{x}{An object.}

\item{...}{Arguments passed on to further methods.}

\item{tz}{The time zone in which to interpret a POSIXt time for extracting
the time of day.  The default is now the zone of \code{x} but was \code{"UTC"}
for v0.3 and earlier.  The previous behavior can be restored by calling
\code{pkgconfig::set_config("hms::default_tz", "UTC")}, see
\code{\link[pkgconfig:set_config]{pkgconfig::set_config()}}.}
}
\description{
\code{is.hms()} has been replaced by \code{\link[=is_hms]{is_hms()}}.

\code{as.hms()} has been replaced by \code{\link[=as_hms]{as_hms()}}, which does not have a \code{tz} argument.
Change the timezone before converting if necessary, e.g. using \code{\link[lubridate:with_tz]{lubridate::with_tz()}}.
}
\keyword{internal}
